@@ -18,15 +18,14 @@ from .. import dependencies
 from .. import mesonlib
 import json
 import subprocess
-from ..coredata import MesonException
+from ..mesonlib import MesonException
 
 class InstallData():
-    def __init__(self, source_dir, build_dir, prefix, depfixer):
+    def __init__(self, source_dir, build_dir, prefix):
         self.source_dir = source_dir
         self.build_dir= build_dir
         self.prefix = prefix
         self.targets = []
-        self.depfixer = depfixer
         self.headers = []
         self.man = []
         self.data = []
@@ -203,27 +202,28 @@ class Backend():
                 if c.can_compile(s):
                     return cpp
         for c in self.build.compilers:
-            if c.get_language() != 'vala':
-                return c
+            if c.get_language() == 'vala':
+                continue
+            for s in src:
+                if c.can_compile(s):
+                    return c
         raise RuntimeError('Unreachable code')
 
+    def object_filename_from_source(self, target, source):
+        return source.fname.replace('/', '_').replace('\\', '_') + '.' + self.environment.get_object_suffix()
+
     def determine_ext_objs(self, extobj, proj_dir_to_build_root=''):
         result = []
         targetdir = self.get_target_private_dir(extobj.target)
-        suffix = '.' + self.environment.get_object_suffix()
         for osrc in extobj.srclist:
-            osrc_base = osrc.fname
-            if not self.source_suffix_in_objs:
-                osrc_base = '.'.join(osrc.split('.')[:-1])
             # If extracting in a subproject, the subproject
             # name gets duplicated in the file name.
             pathsegs = osrc.subdir.split(os.sep)
             if pathsegs[0] == 'subprojects':
                 pathsegs = pathsegs[2:]
             fixedpath = os.sep.join(pathsegs)
-            objbase = osrc.fname.replace('/', '_').replace('\\', '_')
-            objname = os.path.join(proj_dir_to_build_root,
-                                   targetdir, os.path.basename(objbase) + suffix)
+            objname = os.path.join(proj_dir_to_build_root, targetdir,
+                                   self.object_filename_from_source(extobj.target, osrc))
             result.append(objname)
         return result
 
@@ -251,8 +251,6 @@ class Backend():
         commands += self.environment.coredata.external_args[compiler.get_language()]
         commands += target.get_extra_args(compiler.get_language())
         commands += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
-        if self.environment.coredata.get_builtin_option('coverage'):
-            commands += compiler.get_coverage_args()
         if self.environment.coredata.get_builtin_option('werror'):
             commands += compiler.get_werror_args()
         if isinstance(target, build.SharedLibrary):
@@ -380,8 +378,21 @@ class Backend():
             exe_arr = exe.get_command()
         return exe_arr
 
+    def replace_extra_args(self, args, genlist):
+        final_args = []
+        for a in args:
+            if a == '@EXTRA_ARGS@':
+                final_args += genlist.get_extra_args()
+            else:
+                final_args.append(a)
+        return final_args
+
     def eval_custom_target_command(self, target, absolute_paths=False):
-        ofilenames = [os.path.join(self.get_target_dir(target), i) for i in target.output]
+        if not absolute_paths:
+            ofilenames = [os.path.join(self.get_target_dir(target), i) for i in target.output]
+        else:
+            ofilenames = [os.path.join(self.environment.get_build_dir(), self.get_target_dir(target), i) \
+                          for i in target.output]
         srcs = []
         outdir = self.get_target_dir(target)
         # Many external programs fail on empty arguments.
