@@ -17,10 +17,10 @@ from .. import environment, mesonlib
 from .. import build
 from .. import mlog
 from .. import dependencies
-from ..mesonlib import File
+from .. import compilers
+from ..mesonlib import File, MesonException
 from .backends import InstallData
 from ..build import InvalidArguments
-from ..coredata import MesonException
 import os, sys, pickle, re
 import subprocess, shutil
 
@@ -128,7 +128,6 @@ class NinjaBackend(backends.Backend):
 
     def __init__(self, build):
         super().__init__(build)
-        self.source_suffix_in_objs = True
         self.ninja_filename = 'build.ninja'
         self.fortran_deps = {}
         self.all_outputs = {}
@@ -136,7 +135,10 @@ class NinjaBackend(backends.Backend):
     def detect_vs_dep_prefix(self, outfile, tempfilename):
         '''VS writes its dependency in a locale dependent format.
         Detect the search prefix to use.'''
-        if shutil.which('cl') is None:
+        # Of course there is another program called 'cl' on
+        # some platforms. Let's just require that on Windows
+        # cl points to msvc.
+        if not mesonlib.is_windows() or shutil.which('cl') is None:
             return outfile
         outfile.close()
         open(os.path.join(self.environment.get_scratch_dir(), 'incdetect.c'),
@@ -181,7 +183,8 @@ int dummy;
         self.generate_tests(outfile)
         outfile.write('# Install rules\n\n')
         self.generate_install(outfile)
-        if self.environment.coredata.get_builtin_option('coverage'):
+        if 'b_coverage' in self.environment.coredata.base_options and \
+            self.environment.coredata.base_options['b_coverage'].value:
             outfile.write('# Coverage rules\n\n')
             self.generate_coverage_rules(outfile)
         outfile.write('# Suffix\n\n')
@@ -233,7 +236,8 @@ int dummy;
             self.generate_cs_target(target, outfile)
             return
         if 'vala' in self.environment.coredata.compilers.keys() and self.has_vala(target):
-            vala_output_files = self.generate_vala_compile(target, outfile)
+            vc = self.environment.coredata.compilers['vala']
+            vala_output_files = self.generate_vala_compile(vc, target, outfile)
             gen_src_deps += vala_output_files
         if 'swift' in self.environment.coredata.compilers.keys() and self.has_swift(target):
             self.generate_swift_target(target, outfile)
@@ -246,7 +250,7 @@ int dummy;
         self.generate_custom_generator_rules(target, outfile)
         outname = self.get_target_filename(target)
         obj_list = []
-        use_pch = self.environment.coredata.get_builtin_option('use_pch')
+        use_pch = self.environment.coredata.base_options.get('b_pch', False)
         is_unity = self.environment.coredata.get_builtin_option('unity')
         if use_pch and target.has_pch():
             pch_objects = self.generate_pch(target, outfile)
@@ -293,20 +297,20 @@ int dummy;
                                                                          header_deps=header_deps))
         src_list = []
         for src in gen_src_deps:
-                src_list.append(src)
-                if is_unity:
-                    unity_src.append(os.path.join(self.environment.get_build_dir(), src))
+            src_list.append(src)
+            if is_unity:
+                unity_src.append(os.path.join(self.environment.get_build_dir(), src))
+                header_deps.append(src)
+            else:
+                # Generated targets are ordered deps because the must exist
+                # before the sources compiling them are used. After the first
+                # compile we get precise dependency info from dep files.
+                # This should work in all cases. If it does not, then just
+                # move them from orderdeps to proper deps.
+                if self.environment.is_header(src):
                     header_deps.append(src)
                 else:
-                    # Generated targets are ordered deps because the must exist
-                    # before the sources compiling them are used. After the first
-                    # compile we get precise dependency info from dep files.
-                    # This should work in all cases. If it does not, then just
-                    # move them from orderdeps to proper deps.
-                    if self.environment.is_header(src):
-                        header_deps.append(src)
-                    else:
-                        obj_list.append(self.generate_single_compile(target, outfile, src, True, [], header_deps))
+                    obj_list.append(self.generate_single_compile(target, outfile, src, True, [], header_deps))
         for src in target.get_sources():
             if src.endswith('.vala'):
                 continue
@@ -442,10 +446,11 @@ int dummy;
             phony_elem.write(outfile)
 
             elem = NinjaBuildElement(self.all_outputs, 'coveragereport/index.html', 'CUSTOM_COMMAND', '')
+            htmloutdir = os.path.join(self.environment.get_log_dir(), 'coveragereport')
             command = [lcov_exe, '--directory', self.environment.get_build_dir(),\
                        '--capture', '--output-file', 'coverage.info', '--no-checksum',\
                        '&&', genhtml_exe, '--prefix', self.environment.get_build_dir(),\
-                       '--output-directory', self.environment.get_log_dir(), '--title', 'Code coverage',\
+                       '--output-directory', htmloutdir, '--title', 'Code coverage',\
                        '--legend', '--show-details', 'coverage.info']
             elem.add_item('COMMAND', command)
             elem.add_item('DESC', 'Generating HTML coverage report.')
@@ -457,10 +462,9 @@ int dummy;
         script_root = self.environment.get_script_dir()
         install_script = os.path.join(script_root, 'meson_install.py')
         install_data_file = os.path.join(self.environment.get_scratch_dir(), 'install.dat')
-        depfixer = [sys.executable, self.environment.get_build_command(), '--internal', 'depfixer']
         d = InstallData(self.environment.get_source_dir(),
                         self.environment.get_build_dir(),
-                        self.environment.get_prefix(), depfixer)
+                        self.environment.get_prefix())
         elem = NinjaBuildElement(self.all_outputs, 'install', 'CUSTOM_COMMAND', 'PHONY')
         elem.add_dep('all')
         elem.add_item('DESC', 'Installing files.')
@@ -543,12 +547,13 @@ int dummy;
             assert(isinstance(de, build.Data))
             subdir = de.install_dir
             for f in de.sources:
+                plain_f = os.path.split(f)[1]
                 if de.in_sourcetree:
                     srcprefix = self.environment.get_source_dir()
                 else:
                     srcprefix = self.environment.get_build_dir()
                 srcabs = os.path.join(srcprefix, de.source_subdir, f)
-                dstabs = os.path.join(subdir, f)
+                dstabs = os.path.join(subdir, plain_f)
                 i = [srcabs, dstabs]
                 d.data.append(i)
 
@@ -581,7 +586,12 @@ int dummy;
         valgrind = environment.find_valgrind()
         script_root = self.environment.get_script_dir()
         test_data = os.path.join(self.environment.get_scratch_dir(), 'meson_test_setup.dat')
-        cmd = [sys.executable, self.environment.get_build_command(), '--internal', 'test', test_data]
+        cmd = [ sys.executable, self.environment.get_build_command(), '--internal', 'test' ]
+        if not self.environment.coredata.get_builtin_option('stdsplit'):
+            cmd += ['--no-stdsplit']
+        if self.environment.coredata.get_builtin_option('errorlogs'):
+            cmd += ['--print-errorlogs']
+        cmd += [ test_data ]
         elem = NinjaBuildElement(self.all_outputs, 'test', 'CUSTOM_COMMAND', ['all', 'PHONY'])
         elem.add_item('COMMAND', cmd)
         elem.add_item('DESC', 'Running all tests.')
@@ -788,7 +798,7 @@ int dummy;
                     break
         return result
 
-    def generate_vala_compile(self, target, outfile):
+    def generate_vala_compile(self, compiler, target, outfile):
         """Vala is compiled into C. Set up all necessary build steps here."""
         valac = self.environment.coredata.compilers['vala']
         (src, vapi_src) = self.split_vala_sources(target.get_sources())
@@ -808,7 +818,10 @@ int dummy;
 
         generated_c_files = []
         outputs = [vapiname]
-        args = ['-d', self.get_target_private_dir(target)]
+        args = []
+        args += self.build.get_global_args(compiler)
+        args += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
+        args += ['-d', self.get_target_private_dir(target)]
         args += ['-C']#, '-o', cname]
         if not isinstance(target, build.Executable):
             outputs.append(hname)
@@ -1348,13 +1361,7 @@ rule FORTRAN_DEP_HACK
                 relout = self.get_target_private_dir(target)
                 args = [x.replace("@SOURCE_DIR@", self.build_to_src).replace("@BUILD_DIR@", relout)
                         for x in args]
-                final_args = []
-                for a in args:
-                    if a == '@EXTRA_ARGS@':
-                        final_args += genlist.get_extra_args()
-                    else:
-                        final_args.append(a)
-                cmdlist = exe_arr + final_args
+                cmdlist = exe_arr + self.replace_extra_args(args, genlist)
                 elem = NinjaBuildElement(self.all_outputs, outfiles, 'CUSTOM_COMMAND', infilename)
                 if len(extra_dependencies) > 0:
                     elem.add_dep(extra_dependencies)
@@ -1430,6 +1437,8 @@ rule FORTRAN_DEP_HACK
         commands = []
         # The first thing is implicit include directories: source, build and private.
         commands += compiler.get_include_args(self.get_target_private_dir(target), False)
+        commands += compilers.get_base_compile_args(self.environment.coredata.base_options,
+                                                    compiler)
         curdir = target.get_subdir()
         tmppath = os.path.normpath(os.path.join(self.build_to_src, curdir))
         commands += compiler.get_include_args(tmppath, False)
@@ -1484,7 +1493,7 @@ rule FORTRAN_DEP_HACK
         rel_obj = os.path.join(self.get_target_private_dir(target), obj_basename)
         rel_obj += '.' + self.environment.get_object_suffix()
         dep_file = compiler.depfile_for_object(rel_obj)
-        if self.environment.coredata.get_builtin_option('use_pch'):
+        if self.environment.coredata.base_options.get('b_pch', False):
             pchlist = target.get_pch(compiler.language)
         else:
             pchlist = []
@@ -1503,7 +1512,7 @@ rule FORTRAN_DEP_HACK
                     custom_target_include_dirs.append(idir)
         for i in custom_target_include_dirs:
             commands+= compiler.get_include_args(i, False)
-        if self.environment.coredata.get_builtin_option('use_pch'):
+        if self.environment.coredata.base_options.get('b_pch', False):
             commands += self.get_pch_include_args(compiler, target)
         crstr = ''
         if target.is_cross:
@@ -1637,6 +1646,9 @@ rule FORTRAN_DEP_HACK
         abspath = os.path.join(self.environment.get_build_dir(), target.subdir)
         commands = []
         commands += linker.get_linker_always_args()
+        if not isinstance(target, build.StaticLibrary):
+            commands += compilers.get_base_link_args(self.environment.coredata.base_options,
+                                                     linker)
         commands += linker.get_buildtype_linker_args(self.environment.coredata.get_builtin_option('buildtype'))
         commands += linker.get_option_link_args(self.environment.coredata.compiler_options)
         if not(isinstance(target, build.StaticLibrary)):
@@ -1651,6 +1663,8 @@ rule FORTRAN_DEP_HACK
             else:
                 soversion = None
             commands += linker.get_soname_args(target.name, abspath, soversion)
+            if target.vs_module_defs and hasattr(linker, 'gen_vs_module_defs_args'):
+                commands += linker.gen_vs_module_defs_args(target.vs_module_defs.rel_to_builddir(self.build_to_src))
         elif isinstance(target, build.StaticLibrary):
             commands += linker.get_std_link_args()
         else:
@@ -1678,8 +1692,6 @@ rule FORTRAN_DEP_HACK
                         commands += dep.get_link_args()
         commands += linker.build_rpath_args(self.environment.get_build_dir(),\
                                             self.determine_rpath_dirs(target), target.install_rpath)
-        if self.environment.coredata.get_builtin_option('coverage'):
-            commands += linker.get_coverage_link_args()
         custom_target_libraries = self.get_custom_target_provided_libraries(target)
         commands += extra_args
         commands += custom_target_libraries
@@ -1793,7 +1805,8 @@ rule FORTRAN_DEP_HACK
         elem = NinjaBuildElement(self.all_outputs, 'clean', 'CUSTOM_COMMAND', 'PHONY')
         elem.add_item('COMMAND', [ninja_command, '-t', 'clean'])
         elem.add_item('description', 'Cleaning')
-        if self.environment.coredata.get_builtin_option('coverage'):
+        if 'b_coverage' in self.environment.coredata.base_options and \
+           self.environment.coredata.base_options['b_coverage'].value:
             self.generate_gcov_clean(outfile)
             elem.add_dep('clean-gcda')
             elem.add_dep('clean-gcno')
