@@ -14,15 +14,18 @@
 
 import os, sys
 import pickle
+import re
 
 from mesonbuild import compilers
+from mesonbuild.build import BuildTarget
+from mesonbuild.mesonlib import File
 from . import backends
 from .. import build
 from .. import dependencies
 from .. import mlog
 import xml.etree.ElementTree as ET
 import xml.dom.minidom
-from ..coredata import MesonException
+from ..mesonlib import MesonException
 from ..environment import Environment
 
 class RegenInfo():
@@ -35,8 +38,35 @@ class Vs2010Backend(backends.Backend):
     def __init__(self, build):
         super().__init__(build)
         self.project_file_version = '10.0.30319.1'
-        # foo.c compiles to foo.obj, not foo.c.obj
-        self.source_suffix_in_obj = False
+        self.sources_conflicts = {}
+
+    def object_filename_from_source(self, target, source):
+        basename = os.path.basename(source.fname)
+        filename_without_extension = '.'.join(basename.split('.')[:-1])
+        if basename in self.sources_conflicts[target.get_id()]:
+            # If there are multiple source files with the same basename, we must resolve the conflict
+            # by giving each a unique object output file.
+            filename_without_extension = '.'.join(source.fname.split('.')[:-1]).replace('/', '_').replace('\\', '_')
+        return filename_without_extension + '.' + self.environment.get_object_suffix()
+
+    def resolve_source_conflicts(self):
+        for name, target in self.build.targets.items():
+            if not isinstance(target, BuildTarget):
+                continue
+            conflicts = {}
+            for s in target.get_sources():
+                if hasattr(s, 'held_object'):
+                    s = s.held_object
+                if not isinstance(s, File):
+                    continue
+                basename = os.path.basename(s.fname)
+                conflicting_sources = conflicts.get(basename, None)
+                if conflicting_sources is None:
+                    conflicting_sources = []
+                    conflicts[basename] = conflicting_sources
+                conflicting_sources.append(s)
+            self.sources_conflicts[target.get_id()] = {name: src_conflicts for name, src_conflicts in conflicts.items()
+                                                       if len(src_conflicts) > 1}
 
     def generate_custom_generator_commands(self, target, parent_node):
         all_output_files = []
@@ -51,26 +81,24 @@ class Vs2010Backend(backends.Backend):
                 exe = generator.get_exe()
                 infilelist = genlist.get_infilelist()
                 outfilelist = genlist.get_outfilelist()
-                if isinstance(exe, build.BuildTarget):
-                    exe_file = os.path.join(self.environment.get_build_dir(), self.get_target_filename(exe))
-                else:
-                    exe_file = exe.get_command()[0]
+                exe_arr = self.exe_object_to_cmd_array(exe)
                 base_args = generator.get_arglist()
+                target_private_dir = self.relpath(self.get_target_private_dir(target), self.get_target_dir(target))
                 for i in range(len(infilelist)):
                     if len(infilelist) == len(outfilelist):
-                        sole_output = os.path.join(self.get_target_private_dir(target), outfilelist[i])
+                        sole_output = os.path.join(target_private_dir, outfilelist[i])
                     else:
                         sole_output = ''
                     curfile = infilelist[i]
                     infilename = os.path.join(self.environment.get_source_dir(), curfile)
                     outfiles = genlist.get_outputs_for(curfile)
-                    outfiles = [os.path.join(self.get_target_private_dir(target), of) for of in outfiles]
+                    outfiles = [os.path.join(target_private_dir, of) for of in outfiles]
                     all_output_files += outfiles
                     args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)\
                             for x in base_args]
-                    args = [x.replace("@SOURCE_DIR@", self.environment.get_source_dir()).replace("@BUILD_DIR@", self.get_target_private_dir(target))
+                    args = [x.replace("@SOURCE_DIR@", self.environment.get_source_dir()).replace("@BUILD_DIR@", target_private_dir)
                             for x in args]
-                    fullcmd = [exe_file] + args
+                    fullcmd = exe_arr + self.replace_extra_args(args, genlist)
                     commands.append(' '.join(self.special_quote(fullcmd)))
                     inputs.append(infilename)
                     outputs.extend(outfiles)
@@ -86,6 +114,7 @@ class Vs2010Backend(backends.Backend):
         return all_output_files
 
     def generate(self, interp):
+        self.resolve_source_conflicts()
         self.interpreter = interp
         self.platform = 'Win32'
         self.buildtype = self.environment.coredata.get_builtin_option('buildtype')
@@ -116,14 +145,14 @@ class Vs2010Backend(backends.Backend):
         result = {}
         for o in obj_list:
             if isinstance(o, build.ExtractedObjects):
-                result[o.target.get_basename()] = True
+                result[o.target.get_id()] = True
         return result.keys()
 
     def determine_deps(self, p):
         all_deps = {}
         target = self.build.targets[p[0]]
         if isinstance(target, build.CustomTarget):
-            for d in target.dependencies:
+            for d in target.get_target_dependencies():
                 all_deps[d.get_id()] = True
             return all_deps
         if isinstance(target, build.RunTarget):
@@ -220,14 +249,21 @@ class Vs2010Backend(backends.Backend):
         sources = []
         headers = []
         objects = []
+        languages = []
         for i in srclist:
             if self.environment.is_header(i):
                 headers.append(i)
             elif self.environment.is_object(i):
                 objects.append(i)
-            else:
+            elif self.environment.is_source(i):
                 sources.append(i)
-        return (sources, headers, objects)
+                lang = self.lang_from_source_file(i)
+                if lang not in languages:
+                    languages.append(lang)
+            else:
+                # Everything that is not an object or source file is considered a header.
+                headers.append(i)
+        return (sources, headers, objects, languages)
 
     def target_to_build_root(self, target):
         if target.subdir == '':
@@ -272,7 +308,7 @@ class Vs2010Backend(backends.Backend):
         outdir = ET.SubElement(direlem, 'OutDir')
         outdir.text = '.\\'
         intdir = ET.SubElement(direlem, 'IntDir')
-        intdir.text = 'test-temp\\'
+        intdir.text = target.get_id() + '\\'
         tname = ET.SubElement(direlem, 'TargetName')
         tname.text = target.name
         return root
@@ -306,24 +342,26 @@ class Vs2010Backend(backends.Backend):
         (srcs, ofilenames, cmd) = self.eval_custom_target_command(target, True)
         cmd_templ = '''"%s" '''*len(cmd)
         ET.SubElement(customstep, 'Command').text = cmd_templ % tuple(cmd)
-        ET.SubElement(customstep, 'Outputs').text = ';'.join([os.path.join(self.environment.get_build_dir(), i)\
-                                                              for i in ofilenames])
-        ET.SubElement(customstep, 'Inputs').text = ';'.join([os.path.join(self.environment.get_build_dir(), i) \
-                                                             for i in srcs])
+        ET.SubElement(customstep, 'Outputs').text = ';'.join(ofilenames)
+        ET.SubElement(customstep, 'Inputs').text = ';'.join(srcs)
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.targets')
         tree = ET.ElementTree(root)
         tree.write(ofname, encoding='utf-8', xml_declaration=True)
 
+    @classmethod
+    def lang_from_source_file(cls, src):
+        ext = src.split('.')[-1]
+        if ext in compilers.c_suffixes:
+            return 'c'
+        if ext in compilers.cpp_suffixes:
+            return 'cpp'
+        raise MesonException('Could not guess language from source file %s.' % src)
+
     def add_pch(self, inc_cl, proj_to_src_dir, pch_sources, source_file):
         if len(pch_sources) <= 1:
             # We only need per file precompiled headers if we have more than 1 language.
             return
-        if source_file.split('.')[-1] in compilers.c_suffixes:
-            lang = 'c'
-        elif source_file.split('.')[-1] in compilers.cpp_suffixes:
-            lang = 'cpp'
-        else:
-            return
+        lang = Vs2010Backend.lang_from_source_file(source_file)
         header = os.path.join(proj_to_src_dir, pch_sources[lang][0])
         pch_file = ET.SubElement(inc_cl, 'PrecompiledHeaderFile')
         pch_file.text = header
@@ -332,6 +370,17 @@ class Vs2010Backend(backends.Backend):
         pch_out = ET.SubElement(inc_cl, 'PrecompiledHeaderOutputFile')
         pch_out.text = '$(IntDir)$(TargetName)-%s.pch' % lang
 
+    def add_additional_options(self, source_file, parent_node, extra_args, has_additional_options_set):
+        if has_additional_options_set:
+            # We only need per file options if they were not set per project.
+            return
+        lang = Vs2010Backend.lang_from_source_file(source_file)
+        ET.SubElement(parent_node, "AdditionalOptions").text = ' '.join(extra_args[lang]) + ' %(AdditionalOptions)'
+
+    @classmethod
+    def quote_define_cmdline(cls, arg):
+        return re.sub(r'^([-/])D(.*?)="(.*)"$', r'\1D\2=\"\3\"', arg)
+
     def gen_vcxproj(self, target, ofname, guid, compiler):
         mlog.debug('Generating vcxproj %s.' % target.name)
         entrypoint = 'WinMainCRTStartup'
@@ -355,7 +404,7 @@ class Vs2010Backend(backends.Backend):
         down = self.target_to_build_root(target)
         proj_to_src_root = os.path.join(down, self.build_to_src)
         proj_to_src_dir = os.path.join(proj_to_src_root, target.subdir)
-        (sources, headers, objects) = self.split_sources(target.sources)
+        (sources, headers, objects, languages) = self.split_sources(target.sources)
         buildtype = self.buildtype
         project_name = target.name
         target_name = target.name
@@ -388,14 +437,14 @@ class Vs2010Backend(backends.Backend):
         ET.SubElement(type_config, 'UseDebugLibraries').text = 'true'
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.props')
         generated_files = self.generate_custom_generator_commands(target, root)
-        (gen_src, gen_hdrs, gen_objs) = self.split_sources(generated_files)
+        (gen_src, gen_hdrs, gen_objs, gen_langs) = self.split_sources(generated_files)
         direlem = ET.SubElement(root, 'PropertyGroup')
         fver = ET.SubElement(direlem, '_ProjectFileVersion')
         fver.text = self.project_file_version
         outdir = ET.SubElement(direlem, 'OutDir')
         outdir.text = '.\\'
         intdir = ET.SubElement(direlem, 'IntDir')
-        intdir.text = os.path.join(self.get_target_dir(target), target.get_basename() + '.dir') + '\\'
+        intdir.text = target.get_id() + '\\'
         tname = ET.SubElement(direlem, 'TargetName')
         tname.text = target_name
         inclinc = ET.SubElement(direlem, 'LinkIncremental')
@@ -405,38 +454,58 @@ class Vs2010Backend(backends.Backend):
         clconf = ET.SubElement(compiles, 'ClCompile')
         opt = ET.SubElement(clconf, 'Optimization')
         opt.text = 'disabled'
-        inc_dirs = [proj_to_src_dir, self.get_target_private_dir(target)]
-        cur_dir = target.subdir
-        if cur_dir == '':
-            cur_dir= '.'
-        inc_dirs.append(cur_dir)
-        extra_args = []
-        # SUCKS, VS can not handle per-language type flags, so just use
-        # them all.
-        extra_args += compiler.get_buildtype_args(self.buildtype)
-        for l in self.environment.coredata.external_args.values():
-            for a in l:
-                extra_args.append(a)
-        for l in self.build.global_args.values():
-            for a in l:
-                extra_args.append(a)
-        for l in target.extra_args.values():
-            for a in l:
-                extra_args.append(a)
+        inc_dirs = ['.', self.relpath(self.get_target_private_dir(target), self.get_target_dir(target)),
+                    proj_to_src_dir]
+
+        extra_args = {'c': [], 'cpp': []}
+        for l, args in self.environment.coredata.external_args.items():
+            if l in extra_args:
+                extra_args[l] += args
+        for l, args in self.build.global_args.items():
+            if l in extra_args:
+                extra_args[l] += args
+        for l, args in target.extra_args.items():
+            if l in extra_args:
+                extra_args[l] += args
+        general_args = compiler.get_buildtype_args(self.buildtype).copy()
         # FIXME all the internal flags of VS (optimization etc) are represented
         # by their own XML elements. In theory we should split all flags to those
         # that have an XML element and those that don't and serialise them
         # properly. This is a crapton of work for no real gain, so just dump them
         # here.
-        extra_args = compiler.get_option_compile_args(self.environment.coredata.compiler_options)
-        if len(extra_args) > 0:
-            extra_args.append('%(AdditionalOptions)')
-            ET.SubElement(clconf, "AdditionalOptions").text = ' '.join(extra_args)
+        general_args += compiler.get_option_compile_args(self.environment.coredata.compiler_options)
+        for d in target.get_external_deps():
+            try:
+                general_args += d.compile_args
+            except AttributeError:
+                pass
+
+        for l, args in extra_args.items():
+            extra_args[l] = [Vs2010Backend.quote_define_cmdline(x) for x in args]
+
+        languages += gen_langs
+        has_language_specific_args = any(l != extra_args['c'] for l in extra_args.values())
+        additional_options_set = False
+        if not has_language_specific_args or len(languages) == 1:
+            if len(languages) == 0:
+                extra_args = []
+            else:
+                extra_args = extra_args[languages[0]]
+            extra_args = general_args + extra_args
+            if len(extra_args) > 0:
+                extra_args.append('%(AdditionalOptions)')
+                ET.SubElement(clconf, "AdditionalOptions").text = ' '.join(extra_args)
+            additional_options_set = True
+
         for d in target.include_dirs:
             for i in d.incdirs:
                 curdir = os.path.join(d.curdir, i)
                 inc_dirs.append(self.relpath(curdir, target.subdir)) # build dir
                 inc_dirs.append(os.path.join(proj_to_src_root, curdir)) # src dir
+            for i in d.get_extra_build_dirs():
+                curdir = os.path.join(d.curdir, i)
+                inc_dirs.append(self.relpath(curdir, target.subdir))  # build dir
+
         inc_dirs.append('%(AdditionalIncludeDirectories)')
         ET.SubElement(clconf, 'AdditionalIncludeDirectories').text = ';'.join(inc_dirs)
         preproc = ET.SubElement(clconf, 'PreprocessorDefinitions')
@@ -487,14 +556,15 @@ class Vs2010Backend(backends.Backend):
             ET.SubElement(link, "AdditionalOptions").text = ' '.join(extra_link_args)
 
         additional_links = []
-        for t in target.link_targets:
+        for t in target.get_dependencies():
             lobj = self.build.targets[t.get_id()]
             rel_path = self.relpath(lobj.subdir, target.subdir)
             linkname = os.path.join(rel_path, lobj.get_import_filename())
             additional_links.append(linkname)
+        additional_objects = []
         for o in self.flatten_object_list(target, down):
             assert(isinstance(o, str))
-            additional_links.append(o)
+            additional_objects.append(o)
         if len(additional_links) > 0:
             additional_links.append('%(AdditionalDependencies)')
             ET.SubElement(link, 'AdditionalDependencies').text = ';'.join(additional_links)
@@ -515,27 +585,32 @@ class Vs2010Backend(backends.Backend):
         targetmachine = ET.SubElement(link, 'TargetMachine')
         targetmachine.text = 'MachineX86'
 
-        if len(headers) + len(gen_hdrs) > 0:
+        extra_files = target.extra_files
+        if len(headers) + len(gen_hdrs) + len(extra_files) > 0:
             inc_hdrs = ET.SubElement(root, 'ItemGroup')
             for h in headers:
                 relpath = h.rel_to_builddir(proj_to_src_root)
                 ET.SubElement(inc_hdrs, 'CLInclude', Include=relpath)
             for h in gen_hdrs:
-                if isinstance(h, str):
-                    relpath = h
-                else:
-                    relpath = h.rel_to_builddir(proj_to_src_root)
-                ET.SubElement(inc_hdrs, 'CLInclude', Include = relpath)
+                ET.SubElement(inc_hdrs, 'CLInclude', Include=h)
+            for h in target.extra_files:
+                relpath = os.path.join(proj_to_src_dir, h)
+                ET.SubElement(inc_hdrs, 'CLInclude', Include=relpath)
+
         if len(sources) + len(gen_src) + len(pch_sources) > 0:
             inc_src = ET.SubElement(root, 'ItemGroup')
             for s in sources:
                 relpath = s.rel_to_builddir(proj_to_src_root)
                 inc_cl = ET.SubElement(inc_src, 'CLCompile', Include=relpath)
                 self.add_pch(inc_cl, proj_to_src_dir, pch_sources, s)
+                self.add_additional_options(s, inc_cl, extra_args, additional_options_set)
+                basename = os.path.basename(s.fname)
+                if basename in self.sources_conflicts[target.get_id()]:
+                    ET.SubElement(inc_cl, 'ObjectFileName').text = "$(IntDir)" + self.object_filename_from_source(target, s)
             for s in gen_src:
-                relpath =  self.relpath(s, target.subdir)
-                inc_cl = ET.SubElement(inc_src, 'CLCompile', Include=relpath)
+                inc_cl = ET.SubElement(inc_src, 'CLCompile', Include=s)
                 self.add_pch(inc_cl, proj_to_src_dir, pch_sources, s)
+                self.add_additional_options(s, inc_cl, extra_args, additional_options_set)
             for lang in pch_sources:
                 header, impl, suffix = pch_sources[lang]
                 relpath = os.path.join(proj_to_src_dir, impl)
@@ -548,14 +623,17 @@ class Vs2010Backend(backends.Backend):
                 # MSBuild searches for the header relative from the implementation, so we have to use
                 # just the file name instead of the relative path to the file.
                 pch_file.text = os.path.split(header)[1]
+                self.add_additional_options(impl, inc_cl, extra_args, additional_options_set)
 
-        if len(objects) > 0:
+        if len(objects) + len(additional_objects) > 0:
             # Do not add gen_objs to project file. Those are automatically used by MSBuild, because they are part of
             # the CustomBuildStep Outputs.
             inc_objs = ET.SubElement(root, 'ItemGroup')
             for s in objects:
                 relpath = s.rel_to_builddir(proj_to_src_root)
                 ET.SubElement(inc_objs, 'Object', Include=relpath)
+            for s in additional_objects:
+                ET.SubElement(inc_objs, 'Object', Include=s)
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.targets')
         # Reference the regen target.
         ig = ET.SubElement(root, 'ItemGroup')
@@ -604,7 +682,7 @@ class Vs2010Backend(backends.Backend):
         outdir = ET.SubElement(direlem, 'OutDir')
         outdir.text = '.\\'
         intdir = ET.SubElement(direlem, 'IntDir')
-        intdir.text = 'test-temp\\'
+        intdir.text = 'regen-temp\\'
         tname = ET.SubElement(direlem, 'TargetName')
         tname.text = project_name
 
