@@ -17,7 +17,7 @@ from . import environment
 from . import dependencies
 from . import mlog
 import copy, os
-from .mesonlib import File, flatten
+from .mesonlib import File, flatten, MesonException
 
 known_basic_kwargs = {'install' : True,
                       'c_pch' : True,
@@ -47,7 +47,7 @@ known_shlib_kwargs.update({'version' : True,
                            'soversion' : True,
                            'name_prefix' : True,
                            'name_suffix' : True,
-                           })
+                           'vs_module_defs' : True})
 
 backslash_explanation = \
 '''Compiler arguments have a backslash "\\" character. This is unfortunately not
@@ -71,7 +71,7 @@ We are fully aware that these are not really usable or pleasant ways to do
 this but it's the best we can do given the way shell quoting works.
 '''
 
-class InvalidArguments(coredata.MesonException):
+class InvalidArguments(MesonException):
     pass
 
 class Build:
@@ -299,10 +299,10 @@ class BuildTarget():
                 srclist = [srclist]
             for src in srclist:
                 if not isinstance(src, str):
-                    raise coredata.MesonException('Extraction arguments must be strings.')
+                    raise MesonException('Extraction arguments must be strings.')
                 src = File(False, self.subdir, src)
                 if src not in self.sources:
-                    raise coredata.MesonException('Tried to extract unknown source %s.' % src)
+                    raise MesonException('Tried to extract unknown source %s.' % src)
                 obj_src.append(src)
         return ExtractedObjects(self, obj_src)
 
@@ -703,6 +703,7 @@ class SharedLibrary(BuildTarget):
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         self.version = None
         self.soversion = None
+        self.vs_module_defs = None
         super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs);
         if len(self.sources) > 0 and self.sources[0].endswith('.cs'):
             prefix = 'lib'
@@ -726,6 +727,12 @@ class SharedLibrary(BuildTarget):
             self.set_version(kwargs['version'])
         if 'soversion' in kwargs:
             self.set_soversion(kwargs['soversion'])
+        if 'vs_module_defs' in kwargs:
+            path = kwargs['vs_module_defs']
+            if (os.path.isabs(path)):
+                self.vs_module_defs = File.from_absolute_file(path)
+            else:
+                self.vs_module_defs = File.from_source_file(environment.source_dir, self.subdir, path)
 
     def check_unknown_kwargs(self, kwargs):
         self.check_unknown_kwargs_int(kwargs, known_shlib_kwargs)
@@ -806,6 +813,16 @@ class CustomTarget:
     def get_id(self):
         return self.name + self.type_suffix()
 
+    def get_target_dependencies(self):
+        deps = self.dependencies[:]
+        deps += self.extra_depends
+        for c in self.sources:
+            if hasattr(c, 'held_object'):
+                c = c.held_object
+            if isinstance(c, BuildTarget) or isinstance(c, CustomTarget):
+                deps.append(c)
+        return deps
+
     def process_kwargs(self, kwargs):
         self.sources = kwargs.get('input', [])
         if not isinstance(self.sources, list):
@@ -832,6 +849,8 @@ class CustomTarget:
             if isinstance(c, str):
                 final_cmd.append(c)
             elif isinstance(c, dependencies.ExternalProgram):
+                if not c.found():
+                    raise InvalidArguments('Tried to use not found external program in a build rule.')
                 final_cmd += c.get_command()
             elif isinstance(c, BuildTarget) or isinstance(c, CustomTarget):
                 self.dependencies.append(c)
