@@ -16,7 +16,7 @@ import subprocess, os.path
 import tempfile
 from .import mesonlib
 from . import mlog
-from .coredata import MesonException
+from .mesonlib import MesonException
 from . import coredata
 
 """This file contains the data files of all compilers Meson knows
@@ -55,7 +55,9 @@ def is_library(fname):
     return suffix in lib_suffixes
 
 gnulike_buildtype_args = {'plain' : [],
-                          'debug' : ['-g'],
+                          # -O0 is passed for improved debugging information with gcc
+                          # See https://github.com/mesonbuild/meson/pull/509
+                          'debug' : ['-O0', '-g'],
                           'debugoptimized' : ['-O2', '-g'],
                           'release' : ['-O3']}
 
@@ -66,6 +68,7 @@ msvc_buildtype_args = {'plain' : [],
 
 gnulike_buildtype_linker_args = {}
 
+
 if mesonlib.is_osx():
     gnulike_buildtype_linker_args.update({'plain' : [],
                                           'debug' : [],
@@ -111,6 +114,96 @@ msvc_winlibs = ['kernel32.lib', 'user32.lib', 'gdi32.lib',
                 'winspool.lib', 'shell32.lib', 'ole32.lib', 'oleaut32.lib',
                 'uuid.lib', 'comdlg32.lib', 'advapi32.lib']
 
+
+base_options = {
+                'b_pch': coredata.UserBooleanOption('b_pch', 'Use precompiled headers', False),
+                'b_lto': coredata.UserBooleanOption('b_lto', 'Use link time optimization', False),
+                'b_sanitize': coredata.UserComboOption('b_sanitize',
+                                                       'Code sanitizer to use',
+                                                       ['none', 'address', 'thread', 'undefined', 'memory'],
+                                                       'none'),
+                'b_lundef': coredata.UserBooleanOption('b_lundef', 'Use -Wl,--no-undefined when linking', True),
+                'b_pgo': coredata.UserComboOption('b_pgo', 'Use profile guide optimization',
+                                                  ['off', 'generate', 'use'],
+                                                  'off'),
+                'b_coverage': coredata.UserBooleanOption('b_coverage',
+                                                         'Enable coverage tracking.',
+                                                         False),
+                }
+
+def sanitizer_compile_args(value):
+    if value == 'none':
+        return []
+    args = ['-fsanitize=' + value]
+    if value == 'address':
+        args.append('-fno-omit-frame-pointer')
+    return args
+
+def sanitizer_link_args(value):
+    if value == 'none':
+        return []
+    args = ['-fsanitize=' + value]
+    return args
+
+def get_base_compile_args(options, compiler):
+    args = []
+    # FIXME, gcc/clang specific.
+    try:
+        if options['b_lto'].value:
+            args.append('-flto')
+    except KeyError:
+        pass
+    try:
+        args += sanitizer_compile_args(options['b_sanitize'].value)
+    except KeyError:
+        pass
+    try:
+        pgo_val = options['b_pgo'].value
+        if pgo_val == 'generate':
+            args.append('-fprofile-generate')
+        elif pgo_val == 'use':
+            args.append('-fprofile-use')
+    except KeyError:
+        pass
+    try:
+        if options['b_coverage'].value:
+            args += compiler.get_coverage_args()
+    except KeyError:
+        pass
+    return args
+
+def get_base_link_args(options, linker):
+    args = []
+    # FIXME, gcc/clang specific.
+    try:
+        if options['b_lto'].value:
+            args.append('-flto')
+    except KeyError:
+        pass
+    try:
+        args += sanitizer_link_args(options['b_sanitize'].value)
+    except KeyError:
+        pass
+    try:
+        pgo_val = options['b_pgo'].value
+        if pgo_val == 'generate':
+            args.append('-fprofile-generate')
+        elif pgo_val == 'use':
+            args.append('-fprofile-use')
+    except KeyError:
+        pass
+    try:
+        if options['b_lundef'].value:
+            args.append('-Wl,--no-undefined')
+    except KeyError:
+        pass
+    try:
+        if options['b_coverage'].value:
+            args += linker.get_coverage_link_args()
+    except KeyError:
+        pass
+    return args
+
 def build_unix_rpath_args(build_dir, rpath_paths, install_rpath):
         if len(rpath_paths) == 0 and len(install_rpath) == 0:
             return []
@@ -147,6 +240,7 @@ class Compiler():
         else:
             raise TypeError('Unknown argument to Compiler')
         self.version = version
+        self.base_options = []
 
     def get_always_args(self):
         return []
@@ -166,6 +260,9 @@ class Compiler():
     def has_header(self, *args, **kwargs):
         raise EnvironmentException('Language %s does not support header checks.' % self.language)
 
+    def has_header_symbol(self, *args, **kwargs):
+        raise EnvironmentException('Language %s does not support header symbol checks.' % self.language)
+
     def compiles(self, *args, **kwargs):
         raise EnvironmentException('Language %s does not support compile checks.' % self.language)
 
@@ -190,9 +287,12 @@ class Compiler():
     def unix_compile_flags_to_native(self, args):
         return args
 
-    def find_library(self, libname):
+    def find_library(self, libname, extra_dirs):
         raise EnvironmentException('Language {} does not support library finding.'.format(self.language))
 
+    def get_library_dirs(self):
+        return []
+
 class CCompiler(Compiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         super().__init__(exelist, version)
@@ -252,6 +352,9 @@ class CCompiler(Compiler):
     def get_compile_only_args(self):
         return ['-c']
 
+    def get_no_optimization_args(self):
+        return ['-O0']
+
     def get_output_args(self, target):
         return ['-o', target]
 
@@ -280,6 +383,16 @@ class CCompiler(Compiler):
     def get_std_shared_lib_link_args(self):
         return ['-shared']
 
+    def get_library_dirs(self):
+        output = subprocess.Popen(self.exelist + ['--print-search-dirs'], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
+        (stdo, _) = output.communicate()
+        stdo = stdo.decode('utf-8')
+        for line in stdo.split('\n'):
+            if line.startswith('libraries:'):
+                libstr = line.split('=', 1)[1]
+                return libstr.split(':')
+        return []
+
     def can_compile(self, filename):
         suffix = filename.split('.')[-1]
         if suffix == 'c' or suffix == 'h':
@@ -298,6 +411,9 @@ class CCompiler(Compiler):
     def get_pch_name(self, header_name):
         return os.path.split(header_name)[-1] + '.' + self.get_pch_suffix()
 
+    def get_linker_search_args(self, dirname):
+        return ['-L'+dirname]
+
     def sanity_check(self, work_dir):
         mlog.debug('Sanity testing C compiler:', ' '.join(self.exelist))
         mlog.debug('Is cross compiler: %s.' % str(self.is_cross))
@@ -352,6 +468,14 @@ int someSymbolHereJustForFun;
 '''
         return self.compiles(templ % hname, extra_args)
 
+    def has_header_symbol(self, hname, symbol, prefix, extra_args=[]):
+        templ = '''{2}
+#include <{0}>
+int main () {{ {1}; }}'''
+        # Pass -O0 to ensure that the symbol isn't optimized away
+        extra_args += self.get_no_optimization_args()
+        return self.compiles(templ.format(hname, symbol, prefix), extra_args)
+
     def compile(self, code, srcname, extra_args=[]):
         commands = self.get_exelist()
         commands.append(srcname)
@@ -389,7 +513,6 @@ int someSymbolHereJustForFun;
         return p.returncode == 0
 
     def links(self, code, extra_args = []):
-        suflen = len(self.default_suffix)
         (fd, srcname) = tempfile.mkstemp(suffix='.'+self.default_suffix)
         os.close(fd)
         (fd, dstname) = tempfile.mkstemp()
@@ -479,7 +602,7 @@ int main(int argc, char **argv) {
 '''
         res = self.run(templ % (prefix, element), extra_args)
         if not res.compiled:
-            raise EnvironmentException('Could not compile sizeof test.')
+            return -1
         if res.returncode != 0:
             raise EnvironmentException('Could not run sizeof test binary.')
         return int(res.stdout)
@@ -530,15 +653,41 @@ int main(int argc, char **argv) {
         return align
 
     def has_function(self, funcname, prefix, env, extra_args=[]):
-        # This fails (returns true) if funcname is a ptr or a variable.
-        # The correct check is a lot more difficult.
-        # Fix this to do that eventually.
-        templ = '''%s
-int main(int argc, char **argv) {
-    void *ptr = (void*)(%s);
-    return 0;
-};
-'''
+        # Define the symbol to something else in case it is defined by the
+        # includes or defines listed by the user `{0}` or by the compiler.
+        # Then, undef the symbol to get rid of it completely.
+        templ = '''
+        #define {1} meson_disable_define_of_{1}
+        {0}
+        #undef {1}
+        '''
+
+        # Override any GCC internal prototype and declare our own definition for
+        # the symbol. Use char because that's unlikely to be an actual return
+        # value for a function which ensures that we override the definition.
+        templ += '''
+        #ifdef __cplusplus
+        extern "C"
+        #endif
+        char {1} ();
+        '''
+
+        # glibc defines functions that are not available on Linux as stubs that
+        # fail with ENOSYS (such as e.g. lchmod). In this case we want to fail
+        # instead of detecting the stub as a valid symbol.
+        templ += '''
+        #if defined __stub_{1} || defined __stub___{1}
+        fail fail fail this function is not going to work
+        #endif
+        '''
+
+        # And finally the actual function call
+        templ += '''
+        int
+        main ()
+        {{
+          return {1} ();
+        }}'''
         varname = 'has function ' + funcname
         varname = varname.replace(' ', '_')
         if self.is_cross:
@@ -547,7 +696,15 @@ int main(int argc, char **argv) {
                 if isinstance(val, bool):
                     return val
                 raise EnvironmentException('Cross variable {0} is not a boolean.'.format(varname))
-        return self.compiles(templ % (prefix, funcname), extra_args)
+        if self.links(templ.format(prefix, funcname), extra_args):
+            return True
+        # Some functions like alloca() are defined as compiler built-ins which
+        # are inlined by the compiler, so test for that instead. Built-ins are
+        # special functions that ignore all includes and defines, so we just
+        # directly try to link via main().
+        # Add -O0 to ensure that the symbol isn't optimized away by the compiler
+        extra_args += self.get_no_optimization_args()
+        return self.links('int main() {{ {0}; }}'.format('__builtin_' + funcname), extra_args)
 
     def has_member(self, typename, membername, prefix, extra_args=[]):
         templ = '''%s
@@ -566,14 +723,29 @@ void bar() {
 '''
         return self.compiles(templ % (prefix, typename), extra_args)
 
-    def find_library(self, libname):
+    def find_library(self, libname, extra_dirs):
+        # First try if we can just add the library as -l.
         code = '''int main(int argc, char **argv) {
     return 0;
 }
         '''
-        linkarg = '-l' + libname
-        if self.links(code, extra_args=[linkarg]):
-            return linkarg
+        # Gcc + co seem to prefer builtin lib dirs to -L dirs.
+        # Only try to find std libs if no extra dirs specified.
+        if len(extra_dirs) == 0:
+            args = ['-l' + libname]
+            if self.links(code, extra_args=args):
+                return args
+        # Not found? Try to find the library file itself.
+        extra_dirs += self.get_library_dirs()
+        suffixes = ['so', 'dylib', 'lib', 'dll', 'a']
+        for d in extra_dirs:
+            for suffix in suffixes:
+                trial = os.path.join(d, 'lib' + libname + '.' + suffix)
+                if os.path.isfile(trial):
+                    return trial
+                trial2 = os.path.join(d, libname + '.' + suffix)
+                if os.path.isfile(trial2):
+                    return trial2
         return None
 
     def thread_flags(self):
@@ -946,6 +1118,11 @@ class ValaCompiler(Compiler):
         suffix = filename.split('.')[-1]
         return suffix in ('vala', 'vapi')
 
+    def get_buildtype_args(self, buildtype):
+        if buildtype == 'debug' or buildtype == 'debugoptimized':
+            return ['--debug']
+        return []
+
 class RustCompiler(Compiler):
     def __init__(self, exelist, version):
         super().__init__(exelist, version)
@@ -1100,6 +1277,7 @@ class VisualStudioCCompiler(CCompiler):
         self.warn_args = {'1': ['/W2'],
                           '2': ['/W3'],
                           '3': ['/w4']}
+        self.base_options = ['b_pch'] # FIXME add lto, pgo and the like
 
     def get_always_args(self):
         return self.always_args
@@ -1127,6 +1305,9 @@ class VisualStudioCCompiler(CCompiler):
     def get_compile_only_args(self):
         return ['/c']
 
+    def get_no_optimization_args(self):
+        return ['/Od']
+
     def get_output_args(self, target):
         if target.endswith('.exe'):
             return ['/Fe' + target]
@@ -1144,12 +1325,22 @@ class VisualStudioCCompiler(CCompiler):
     def get_linker_output_args(self, outputname):
         return ['/OUT:' + outputname]
 
+    def get_linker_search_args(self, dirname):
+        return ['/LIBPATH:' + dirname]
+
     def get_pic_args(self):
         return [] # PIC is handled by the loader on Windows
 
     def get_std_shared_lib_link_args(self):
         return ['/DLL']
 
+    def gen_vs_module_defs_args(self, defsfile):
+        if not isinstance(defsfile, str):
+            raise RuntimeError('Module definitions file should be str')
+        # With MSVC, DLLs only export symbols that are explicitly exported,
+        # so if a module defs file is specified, we use that to export symbols
+        return ['/DEF:' + defsfile]
+
     def gen_pch_args(self, header, source, pchname):
         objname = os.path.splitext(pchname)[0] + '.obj'
         return (objname, ['/Yc' + header, '/Fp' + pchname, '/Fo' + objname ])
@@ -1175,6 +1366,19 @@ class VisualStudioCCompiler(CCompiler):
     def build_rpath_args(self, build_dir, rpath_paths, install_rpath):
         return []
 
+    def find_library(self, libname, extra_dirs):
+        code = '''int main(int argc, char **argv) {
+    return 0;
+}
+        '''
+        args = []
+        for i in extra_dirs:
+            args += self.get_linker_search_args(i)
+        args.append(libname + '.lib')
+        if self.links(code, extra_args=args):
+            return args
+        return None
+
     # FIXME, no idea what these should be.
     def thread_flags(self):
         return []
@@ -1198,7 +1402,13 @@ class VisualStudioCCompiler(CCompiler):
                 i = '/LIBPATH:' + i[2:]
             # Translate GNU-style -lfoo library name to the import library
             if i.startswith('-l'):
-                i = i[2:] + '.lib'
+                name = i[2:]
+                if name in ('m', 'c'):
+                    # With MSVC, these are provided by the C runtime which is
+                    # linked in by default
+                    continue
+                else:
+                    i = name + '.lib'
             result.append(i)
         return result
 
@@ -1222,6 +1432,7 @@ class VisualStudioCPPCompiler(VisualStudioCCompiler):
         VisualStudioCCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
         self.language = 'cpp'
         self.default_suffix = 'cpp'
+        self.base_options = ['b_pch'] # FIXME add lto, pgo and the like
 
     def can_compile(self, filename):
         suffix = filename.split('.')[-1]
@@ -1271,6 +1482,11 @@ GCC_STANDARD = 0
 GCC_OSX = 1
 GCC_MINGW = 2
 
+CLANG_STANDARD = 0
+CLANG_OSX = 1
+CLANG_WIN = 2
+# Possibly clang-cl?
+
 def get_gcc_soname_args(gcc_type, shlib_name, path, soversion):
     if soversion is None:
         sostr = ''
@@ -1293,6 +1509,9 @@ class GnuCCompiler(CCompiler):
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch'],
                           '2': ['-Wall', '-Wextra', '-Winvalid-pch'],
                           '3' : ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch']}
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        if self.gcc_type != GCC_OSX:
+            self.base_options.append('b_lundef')
 
     def get_pic_args(self):
         if self.gcc_type == GCC_MINGW:
@@ -1355,6 +1574,9 @@ class GnuObjCCompiler(ObjCCompiler):
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch'],
                           '2': ['-Wall', '-Wextra', '-Winvalid-pch'],
                           '3' : ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch']}
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        if self.gcc_type != GCC_OSX:
+            self.base_options.append('b_lundef')
 
     def get_buildtype_args(self, buildtype):
         return gnulike_buildtype_args[buildtype]
@@ -1380,6 +1602,9 @@ class GnuObjCPPCompiler(ObjCPPCompiler):
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '2': ['-Wall', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '3' : ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor']}
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        if self.gcc_type != GCC_OSX:
+            self.base_options.append('b_lundef')
 
     def get_buildtype_args(self, buildtype):
         return gnulike_buildtype_args[buildtype]
@@ -1394,22 +1619,34 @@ class GnuObjCPPCompiler(ObjCPPCompiler):
         return get_gcc_soname_args(self.gcc_type, shlib_name, path, soversion)
 
 class ClangObjCCompiler(GnuObjCCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+    def __init__(self, exelist, version, cltype, is_cross, exe_wrapper=None):
         super().__init__(exelist, version, is_cross, exe_wrapper)
         self.id = 'clang'
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        self.clang_type = cltype
+        if self.clang_type != CLANG_OSX:
+            self.base_options.append('b_lundef')
 
 class ClangObjCPPCompiler(GnuObjCPPCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+    def __init__(self, exelist, version, cltype, is_cross, exe_wrapper=None):
         super().__init__(exelist, version, is_cross, exe_wrapper)
         self.id = 'clang'
+        self.clang_type = cltype
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        if self.clang_type != CLANG_OSX:
+            self.base_options.append('b_lundef')
 
 class ClangCCompiler(CCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+    def __init__(self, exelist, version, clang_type, is_cross, exe_wrapper=None):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
         self.id = 'clang'
+        self.clang_type = clang_type
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch'],
                           '2': ['-Wall', '-Wextra', '-Winvalid-pch'],
                           '3' : ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch']}
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        if self.clang_type != CLANG_OSX:
+            self.base_options.append('b_lundef')
 
     def get_buildtype_args(self, buildtype):
         return gnulike_buildtype_args[buildtype]
@@ -1455,6 +1692,9 @@ class GnuCPPCompiler(CPPCompiler):
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '2': ['-Wall', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '3': ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor']}
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        if self.gcc_type != GCC_OSX:
+            self.base_options.append('b_lundef')
 
     def get_always_args(self):
         return ['-pipe']
@@ -1495,12 +1735,16 @@ class GnuCPPCompiler(CPPCompiler):
         return []
 
 class ClangCPPCompiler(CPPCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+    def __init__(self, exelist, version, cltype, is_cross, exe_wrapper=None):
         CPPCompiler.__init__(self, exelist, version, is_cross, exe_wrapper)
         self.id = 'clang'
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '2': ['-Wall', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '3': ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor']}
+        self.clang_type = cltype
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        if self.clang_type != CLANG_OSX:
+            self.base_options.append('b_lundef')
 
     def get_buildtype_args(self, buildtype):
         return gnulike_buildtype_args[buildtype]
