@@ -19,7 +19,7 @@ import configparser
 
 build_filename = 'meson.build'
 
-class EnvironmentException(coredata.MesonException):
+class EnvironmentException(mesonlib.MesonException):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
@@ -222,7 +222,11 @@ class Environment():
                     gtype = GCC_STANDARD
                 return GnuCCompiler(ccache + [compiler], version, gtype, is_cross, exe_wrap)
             if 'clang' in out:
-                return ClangCCompiler(ccache + [compiler], version, is_cross, exe_wrap)
+                if 'Apple' in out:
+                    cltype = CLANG_OSX
+                else:
+                    cltype = CLANG_STANDARD
+                return ClangCCompiler(ccache + [compiler], version, cltype, is_cross, exe_wrap)
             if 'Microsoft' in out or 'Microsoft' in err:
                 # Visual Studio prints version number to stderr but
                 # everything else to stdout. Why? Lord only knows.
@@ -247,11 +251,11 @@ class Environment():
         for compiler in compilers:
             for arg in ['--version', '-V']:
                 try:
-                   p = subprocess.Popen([compiler] + [arg],
-                                        stdout=subprocess.PIPE,
-                                        stderr=subprocess.PIPE)
+                    p = subprocess.Popen([compiler] + [arg],
+                                         stdout=subprocess.PIPE,
+                                         stderr=subprocess.PIPE)
                 except OSError:
-                   continue
+                    continue
                 (out, err) = p.communicate()
                 out = out.decode(errors='ignore')
                 err = err.decode(errors='ignore')
@@ -262,32 +266,32 @@ class Environment():
                     version = vmatch.group(0)
 
                 if 'GNU Fortran' in out:
-                  return GnuFortranCompiler([compiler], version, GCC_STANDARD, is_cross, exe_wrap)
+                    return GnuFortranCompiler([compiler], version, GCC_STANDARD, is_cross, exe_wrap)
 
                 if 'G95' in out:
-                  return G95FortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return G95FortranCompiler([compiler], version, is_cross, exe_wrap)
 
                 if 'Sun Fortran' in err:
-                  version = 'unknown version'
-                  vmatch = re.search(Environment.version_regex, err)
-                  if vmatch:
-                      version = vmatch.group(0)
-                  return SunFortranCompiler([compiler], version, is_cross, exe_wrap)
+                    version = 'unknown version'
+                    vmatch = re.search(Environment.version_regex, err)
+                    if vmatch:
+                        version = vmatch.group(0)
+                    return SunFortranCompiler([compiler], version, is_cross, exe_wrap)
 
                 if 'ifort (IFORT)' in out:
-                  return IntelFortranCompiler([compiler], version, is_cross, exe_wrap)
-                
+                    return IntelFortranCompiler([compiler], version, is_cross, exe_wrap)
+
                 if 'PathScale EKOPath(tm)' in err:
-                  return PathScaleFortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return PathScaleFortranCompiler([compiler], version, is_cross, exe_wrap)
 
                 if 'pgf90' in out:
-                  return PGIFortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return PGIFortranCompiler([compiler], version, is_cross, exe_wrap)
 
                 if 'Open64 Compiler Suite' in err:
-                  return Open64FortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return Open64FortranCompiler([compiler], version, is_cross, exe_wrap)
 
                 if 'NAG Fortran' in err:
-                  return NAGFortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return NAGFortranCompiler([compiler], version, is_cross, exe_wrap)
 
         raise EnvironmentException('Unknown compiler(s): "' + ', '.join(compilers) + '"')
 
@@ -346,7 +350,11 @@ class Environment():
                     gtype = GCC_STANDARD
                 return GnuCPPCompiler(ccache + [compiler], version, gtype, is_cross, exe_wrap)
             if 'clang' in out:
-                return ClangCPPCompiler(ccache + [compiler], version, is_cross, exe_wrap)
+                if 'Apple' in out:
+                    cltype = CLANG_OSX
+                else:
+                    cltype = CLANG_STANDARD
+                return ClangCPPCompiler(ccache + [compiler], version, cltype, is_cross, exe_wrap)
             if 'Microsoft' in out or 'Microsoft' in err:
                 version = re.search(Environment.version_regex, err).group()
                 return VisualStudioCPPCompiler([compiler], version, is_cross, exe_wrap)
@@ -377,7 +385,7 @@ class Environment():
             'Free Software Foundation' in out:
             return GnuObjCCompiler(exelist, version, is_cross, exe_wrap)
         if out.startswith('Apple LLVM'):
-            return ClangObjCCompiler(exelist, version, is_cross, exe_wrap)
+            return ClangObjCCompiler(exelist, version, CLANG_OSX, is_cross, exe_wrap)
         if 'apple' in out and 'Free Software Foundation' in out:
             return GnuObjCCompiler(exelist, version, is_cross, exe_wrap)
         raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
@@ -407,7 +415,7 @@ class Environment():
             'Free Software Foundation' in out:
             return GnuObjCPPCompiler(exelist, version, is_cross, exe_wrap)
         if out.startswith('Apple LLVM'):
-            return ClangObjCPPCompiler(exelist, version, is_cross, exe_wrap)
+            return ClangObjCPPCompiler(exelist, version, CLANG_OSX, is_cross, exe_wrap)
         if 'apple' in out and 'Free Software Foundation' in out:
             return GnuObjCPPCompiler(exelist, version, is_cross, exe_wrap)
         raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
@@ -506,7 +514,7 @@ class Environment():
             evar = 'AR'
             if evar in os.environ:
                 linker = os.environ[evar].strip()
-            if isinstance(compiler, VisualStudioCCompiler):
+            elif isinstance(compiler, VisualStudioCCompiler):
                 linker= self.vs_static_linker
             else:
                 linker = self.default_static_linker
@@ -590,6 +598,9 @@ class Environment():
     def get_libdir(self):
         return self.coredata.get_builtin_option('libdir')
 
+    def get_libexecdir(self):
+        return self.coredata.get_builtin_option('libexecdir')
+
     def get_bindir(self):
         return self.coredata.get_builtin_option('bindir')
 
@@ -646,11 +657,11 @@ class CrossBuildInfo():
         if 'target_machine' in self.config:
             return
         if not 'host_machine' in self.config:
-            raise coredata.MesonException('Cross info file must have either host or a target machine.')
+            raise mesonlib.MesonException('Cross info file must have either host or a target machine.')
         if not 'properties' in self.config:
-            raise coredata.MesonException('Cross file is missing "properties".')
+            raise mesonlib.MesonException('Cross file is missing "properties".')
         if not 'binaries' in self.config:
-            raise coredata.MesonException('Cross file is missing "binaries".')
+            raise mesonlib.MesonException('Cross file is missing "binaries".')
 
     def ok_type(self, i):
         return isinstance(i, str) or isinstance(i, int) or isinstance(i, bool)
