@@ -16,6 +16,7 @@
 
 import sys, pickle, os, shutil, subprocess, gzip, platform
 from glob import glob
+from mesonbuild.scripts import depfixer
 
 def do_install(datafilename):
     ifile = open(datafilename, 'rb')
@@ -118,10 +119,14 @@ def run_install_script(d):
         if platform.system().lower() == 'windows' and suffix != '.bat':
             first_line = open(script).readline().strip()
             if first_line.startswith('#!'):
-                commands = first_line[2:].split('#')[0].strip().split()
-                commands[0] = shutil.which(commands[0].split('/')[-1])
-                if commands[0] is None:
-                    raise RuntimeError("Don't know how to run script %s." % script)
+                if shutil.which(first_line[2:]):
+                    commands = [first_line[2:]]
+                else:
+                    commands = first_line[2:].split('#')[0].strip().split()
+                    commands[0] = shutil.which(commands[0].split('/')[-1])
+                    if commands[0] is None:
+                        commands
+                        raise RuntimeError("Don't know how to run script %s." % script)
                 final_command = commands + [script] + i.cmd_arr[1:]
         else:
             final_command = i.cmd_arr
@@ -129,8 +134,8 @@ def run_install_script(d):
             rc = subprocess.call(final_command, env=child_env)
             if rc != 0:
                 sys.exit(rc)
-        except Exception:
-            print('Failed to run install script:', i.cmd_arr[0])
+        except:
+            print('Failed to run install script:', *i.cmd_arr)
             sys.exit(1)
 
 def is_elf_platform():
@@ -199,15 +204,14 @@ def install_targets(d):
                     print("Symlink creation does not work on this platform.")
                     printed_symlink_error = True
         if is_elf_platform():
-            p = subprocess.Popen(d.depfixer + [outname, install_rpath],
-                                 stdout=subprocess.PIPE,
-                                 stderr=subprocess.PIPE)
-            (stdo, stde) = p.communicate()
-            if p.returncode != 0:
-                print('Could not fix dependency info.\n')
-                print('Stdout:\n%s\n' % stdo.decode())
-                print('Stderr:\n%s\n' % stde.decode())
-                sys.exit(1)
+            try:
+                e = depfixer.Elf(outname, False)
+                e.fix_rpath(install_rpath)
+            except SystemExit as e:
+                if isinstance(e.code, int) and e.code == 0:
+                    pass
+                else:
+                    raise
 
 def run(args):
     if len(args) != 1:
